/****************************************************************************
**
** Copyright (C) 2007-2009 Kevin Clague. All rights reserved.
** Copyright (C) 2015 - 2024 Trevor SANDY. All rights reserved.
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  Please review the following information to ensure GNU
** General Public Licensing requirements will be met:
** http://www.trolltech.com/products/qt/opensource.html
* sales department at sales@trolltech.com.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

/****************************************************************************
 *
 * This file describes a dialog the user can use to edit divider attributes
 *
 * Please see lpub.h for an overall description of how the files in LPub
 * make up the LPub program.
 *
 ***************************************************************************/

#include "dividerdialog.h"
#include <QVBoxLayout>
#include <QGroupBox>
#include <QDialogButtonBox>

#include "metagui.h"
#include "commonmenus.h"

DividerDialog::DividerDialog(
  SepData &goods,
  QString  _name,
  QWidget *parent)
  : QDialog(parent)
{
  setWindowTitle(tr("Edit %1").arg(_name));

  setWhatsThis(lpubWT(WT_DIALOG_DIVIDER,windowTitle()));

  meta.setValue(goods);
  QVBoxLayout *layout = new QVBoxLayout(this);
  setLayout(layout);

  QGroupBox *box = new QGroupBox("Divider",this);
  layout->addWidget(box);
  seperator = new SepGui(&meta,box);

  QDialogButtonBox *buttonBox;
  
  buttonBox = new QDialogButtonBox(this);

  buttonBox->addButton(QDialogButtonBox::Ok);
  connect(buttonBox,SIGNAL(accepted()),SLOT(accept()));
  buttonBox->addButton(QDialogButtonBox::Cancel);
  connect(buttonBox,SIGNAL(rejected()),SLOT(cancel()));

  layout->addWidget(buttonBox);

  setModal(true);
  setMinimumSize(40,20);
}

DividerDialog::~DividerDialog()
{
}

bool DividerDialog::getDivider(
  SepData &goods,
  QString  name,
  QWidget *parent)
{
  DividerDialog *dialog = new DividerDialog(goods,name,parent);

  bool ok = dialog->exec() == QDialog::Accepted;
  if (ok) {
    goods = dialog->meta.value();
  }
  return ok;
}

void DividerDialog::accept()
{
  if (seperator->modified) {
    QDialog::accept();
  } else {
    QDialog::reject();
  }
}

void DividerDialog::cancel()
{
  QDialog::reject();
}
